#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
import argparse
import io
import time
from pathlib import Path

from PIL import Image, ImageChops
from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions
from selenium.webdriver.support.ui import WebDriverWait


def to_url(path):
    if path.is_file():
        return path.resolve().as_uri()
    return path


def screenshot(browser, url):
    browser.get(url)

    target_find_by = By.TAG_NAME
    target = "body"
    loaded_page_settling_time = 0

    # Selenium doesn't like when we try to screenshot <body> element of documents generated by pdf2htmlEX
    if "poppler" in url:
        target_find_by = By.ID
        target = "page-container"
        loaded_page_settling_time = 1

    web_driver_wait = WebDriverWait(browser, 5)
    web_driver_wait.until(
        expected_conditions.presence_of_element_located((target_find_by, target))
    )
    web_driver_wait.until(
        lambda driver: driver.execute_script("return document.readyState") == "complete"
    )
    if loaded_page_settling_time != 0:
        time.sleep(loaded_page_settling_time)

    target_element = browser.find_element(target_find_by, target)

    png = target_element.screenshot_as_png
    return Image.open(io.BytesIO(png))


def get_browser(driver="firefox", max_width=1000, max_height=10000):
    if driver == "phantomjs":
        browser = webdriver.PhantomJS()
    elif driver == "firefox":
        options = webdriver.FirefoxOptions()
        options.add_argument("--headless")
        browser = webdriver.Firefox(options=options)
    else:  # chrome or unknown
        options = webdriver.ChromeOptions()
        options.add_argument("--headless=new")
        browser = webdriver.Chrome(options=options)
    browser.set_window_size(max_width, max_height)
    return browser


def html_render_diff(a, b, browser):
    image_a = screenshot(browser, to_url(a))
    image_b = screenshot(browser, to_url(b))

    image_a = image_a.convert("RGB")
    image_b = image_b.convert("RGB")
    diff = ImageChops.difference(image_a, image_b)
    return diff, (image_a, image_b)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("a", type=Path, help="Path to the first HTML file")
    parser.add_argument("b", type=Path, help="Path to the second HTML file")
    parser.add_argument(
        "--driver", choices=["chrome", "firefox", "phantomjs"], default="firefox"
    )
    parser.add_argument("--max-width", default=1000)
    parser.add_argument("--max-height", default=10000)
    args = parser.parse_args()

    browser = get_browser(args.driver, args.max_width, args.max_height)
    diff, _ = html_render_diff(args.a, args.b, browser)
    browser.quit()
    bounding_box = diff.getbbox()

    if bounding_box:
        print("images are different")
        print("first error at %f%%" % (1e2 * bounding_box[1] / diff.height))
        print(
            "bounding box %f%%"
            % (
                1e2
                * (
                    (bounding_box[2] - bounding_box[0])
                    * (bounding_box[3] - bounding_box[1])
                )
                / (diff.width * diff.height)
            )
        )
        return 1

    print("images are the same")
    return 0


if __name__ == "__main__":
    sys.exit(main())
